from typing import Tuple
from torchseal.utils import approximate_toeplitz_multiple_channels, precise_toeplitz_multiple_channels, generate_near_zeros

import numpy as np
import torch


def create_conv2d_weight_mask(input_size_with_channel: Tuple[int, int, int], kernel_size_with_channel: Tuple[int, int, int, int], stride: int = 1, padding: int = 0) -> torch.Tensor:
    # Get the shapes
    out_channels, _, kernel_height, kernel_width = kernel_size_with_channel
    in_channels, input_height, input_width = input_size_with_channel

    # Calculate the length of the index tensor
    index_length = out_channels * in_channels * kernel_height * kernel_width

    # Create the index tensor
    index_tensor = approximate_toeplitz_multiple_channels(
        torch.arange(1, index_length + 1).view(
            out_channels, in_channels, kernel_height, kernel_width
        ),
        (in_channels, input_height, input_width),
        stride=stride,
        padding=padding
    )

    # Create the binary tensor
    binary_tensor = torch.stack([
        torch.tensor(
            np.vectorize(
                lambda x: 1.000 if np.isclose(x, i, atol=1e-3, rtol=0) else generate_near_zeros(
                    (1, 1)
                ).item()
            )(index_tensor)
        )
        for i in range(1, index_length + 1)
    ])

    return binary_tensor


def create_conv2d_bias_transformation(repeat: int, length: int) -> torch.Tensor:
    # Calculate the number of groups
    num_groups = length // repeat

    # Create the transformation matrix
    transformation_matrix = generate_near_zeros((length, length))

    # Fill the transformation matrix
    for i in range(num_groups):
        for j in range(repeat):
            row = i * repeat + j
            group_start = i * repeat
            transformation_matrix[row, group_start:group_start + repeat] = 1

    return transformation_matrix



from typing import Tuple
from torchseal.wrapper import CKKSWrapper
from torchseal.function import AvgPool2dFunction
from torchseal.utils import approximate_toeplitz_multiple_channels, create_average_kernel, create_padding_transformation_matrix, create_inverse_padding_transformation_matrix

import typing
import torch


class AvgPool2d(torch.nn.Module):
    weight: torch.Tensor
    conv2d_padding_transformation: torch.Tensor
    conv2d_inverse_padding_transformation: torch.Tensor

    def __init__(self, n_channels: int, kernel_size: Tuple[int, int], input_size: Tuple[int, int], stride: int = 1, padding: int = 0) -> None:
        super(AvgPool2d, self).__init__()

        # Unpack the kernel size
        kernel_height, kernel_width = kernel_size

        # Unpack the input size
        input_height, input_width = input_size

        # Create the weight matrix for the average pooling operation
        self.weight = approximate_toeplitz_multiple_channels(
            create_average_kernel(
                n_channels, kernel_height, kernel_width
            ),
            (n_channels, input_height, input_width),
            stride=stride,
            padding=padding
        )

        # Create the binary masking for inference
        self.conv2d_padding_transformation = create_padding_transformation_matrix(
            n_channels, input_height, input_width, padding
        )

        self.conv2d_inverse_padding_transformation = create_inverse_padding_transformation_matrix(
            n_channels, input_height, input_width, padding
        )

    def forward(self, enc_x: CKKSWrapper) -> CKKSWrapper:
        enc_output = typing.cast(
            CKKSWrapper,
            AvgPool2dFunction.apply(
                enc_x, self.weight, self.conv2d_padding_transformation, self.conv2d_inverse_padding_transformation
            )
        )

        return enc_output



from typing import Tuple, Optional
from torchseal.wrapper import CKKSWrapper, CKKSLinearFunctionWrapper

import typing
import torch
import torchseal


class Conv2dFunction(torch.autograd.Function):
    @staticmethod
    def forward(ctx: CKKSLinearFunctionWrapper, enc_input: CKKSWrapper, weight: CKKSWrapper, bias: CKKSWrapper, conv2d_padding_transformation: torch.Tensor, conv2d_inverse_padding_transformation: torch.Tensor, conv2d_weight_mask: torch.Tensor, conv2d_bias_transformation: torch.Tensor, training: bool) -> CKKSWrapper:
        # Apply the padding transformation to the encrypted input
        # NOTE: This is useless if padding is 0 (we can skip this step if that's the case)
        enc_padding_input = enc_input.ckks_matrix_multiplication(
            conv2d_padding_transformation
        )

        # Save the ctx for the backward method
        ctx.save_for_backward(
            conv2d_inverse_padding_transformation,
            conv2d_weight_mask,
            conv2d_bias_transformation
        )
        ctx.enc_input = enc_padding_input.clone()
        ctx.weight = weight.clone()

        # Apply the convolution to the encrypted input
        # If training, apply the linear transformation to the encrypted input using encrypted parameters
        if training:
            enc_output = enc_padding_input.ckks_encrypted_matrix_multiplication(
                weight.ckks_transpose()
            ).ckks_encrypted_addition(bias)

            return enc_output

        # Else, apply the linear transformation to the encrypted input using plaintext parameters
        enc_output = enc_padding_input.ckks_matrix_multiplication(
            weight.plaintext_data.t()
        ).ckks_addition(bias.plaintext_data)

        return enc_output

    @staticmethod
    def backward(ctx: CKKSLinearFunctionWrapper, enc_grad_output: CKKSWrapper) -> Tuple[Optional[CKKSWrapper], Optional[CKKSWrapper], Optional[CKKSWrapper], None, None, None, None, None]:
        # Get the saved tensors
        conv2d_inverse_padding_transformation, conv2d_weight_mask, conv2d_bias_transformation = typing.cast(
            Tuple[
                torch.Tensor, torch.Tensor, torch.Tensor
            ],
            ctx.saved_tensors
        )
        enc_input = ctx.enc_input
        enc_weight = ctx.weight

        # Get the needs_input_grad
        result = typing.cast(
            Tuple[bool, bool, bool, bool, bool, bool, bool, bool],
            ctx.needs_input_grad
        )

        # Initialize the gradients
        enc_grad_input = enc_grad_weight = enc_grad_bias = None

        if result[0]:
            # Calculate the gradients for the input tensor (this will be encrypted)
            enc_padded_grad_input = enc_grad_output.ckks_encrypted_matrix_multiplication(
                enc_weight
            )

            # Apply the inverse padding transformation to the gradient input
            # NOTE: This is useless if padding is 0 (we can skip this step if that's the case)
            enc_grad_input = enc_padded_grad_input.ckks_matrix_multiplication(
                conv2d_inverse_padding_transformation
            )

        if result[1]:
            # Create the fully connected gradient weight tensor (this will be encrypted)
            enc_unfiltered_grad_weight = enc_grad_output.ckks_transpose().ckks_encrypted_matrix_multiplication(
                enc_input
            )

            # Initialize the gradient weight tensor (this will be encrypted, probably going to be plain tensor)
            enc_grad_weight = torchseal.ckks_zeros(
                enc_weight.shape, do_encryption=True
            )

            # Apply the binary tensor to the gradient weight
            for binary_mask in conv2d_weight_mask:
                # Apply the binary mask to the gradient weight (this will be encrypted)
                enc_filtered_grad_weight = enc_unfiltered_grad_weight.ckks_apply_mask(
                    binary_mask
                )

                # Calculate the sum of all elements (this will be encrypted)
                enc_sum_all_element = enc_filtered_grad_weight.ckks_sum(
                    axis=1
                ).ckks_sum(
                    axis=0
                )

                # Create the new current gradient weight (this will be encrypted)
                enc_current_gradient_weight = enc_sum_all_element.ckks_apply_mask(
                    binary_mask
                )

                # Add the current gradient weight to the final gradient weight (this will be encrypted)
                enc_grad_weight = enc_grad_weight.ckks_encrypted_addition(
                    enc_current_gradient_weight
                )

        if result[2]:
            # Apply the binary transformation to the gradient output (this will be encrypted)
            enc_linear_grad_bias = enc_grad_output.ckks_sum(
                axis=0
            )

            # Calculate the gradient bias for the convolutional layer (this will be encrypted)
            enc_grad_bias = enc_linear_grad_bias.ckks_apply_transformation(
                conv2d_bias_transformation
            )

        return enc_grad_input, enc_grad_weight, enc_grad_bias, None, None, None, None, None



from typing import Any, Dict, Optional
from threading import Lock

import tenseal as ts


# Source: https://refactoring.guru/design-patterns/singleton/python/example#example-1
# More about metaclasses: https://realpython.com/python-metaclasses/
class CKKSStateMeta(type):
    _instances: Dict["CKKSStateMeta", "CKKSState"] = {}
    _lock: Lock = Lock()

    def __call__(self, *args: Any, **kwargs: Any) -> "CKKSState":
        # Now, imagine that the program has just been launched. Since there's no
        # Singleton instance yet, multiple threads can simultaneously pass the
        # previous conditional and reach this point almost at the same time. The
        # first of them will acquire lock and will proceed further, while the
        # rest will wait here.

        with self._lock:
            # The first thread to acquire the lock, reaches this conditional,
            # goes inside and creates the Singleton instance. Once it leaves the
            # lock block, a thread that might have been waiting for the lock
            # release may then enter this section. But since the Singleton field
            # is already initialized, the thread won't create a new object.

            if self not in self._instances:
                instance = super().__call__(*args, **kwargs)
                self._instances[self] = instance

        return self._instances[self]


# Source: https://refactoring.guru/design-patterns/singleton/python/example#example-1
class CKKSState(metaclass=CKKSStateMeta):
    __context: Optional[ts.Context] = None

    @property
    def context(self) -> ts.Context:
        assert self.__context is not None, "Context is not set"

        return self.__context

    @context.setter
    def context(self, context: ts.Context) -> None:
        self.__context = context



from typing import Literal, Union
from numpy.polynomial import Polynomial, Chebyshev
from torchseal.wrapper import CKKSWrapper
from torchseal.function import TanhFunction

import typing
import numpy as np
import torch


class Tanh(torch.nn.Module):
    coeffs: np.ndarray

    def __init__(self, start: float, stop: float, num_of_sample: int, degree: int, approximation_type: Union[Literal["minimax"], Literal["least-squares"]]) -> None:
        super(Tanh, self).__init__()

        # Create the polynomial
        x = np.linspace(start, stop, num_of_sample)
        y = (lambda x: np.tanh(x))(x)

        # Perform the polynomial approximation
        self.coeffs = Polynomial.fit(
            x, y, degree
        ).convert(kind=Polynomial).coef if approximation_type == "least-squares" else Chebyshev.fit(x, y, degree).convert(kind=Polynomial).coef

    def forward(self, enc_x: CKKSWrapper) -> CKKSWrapper:
        enc_output = typing.cast(
            CKKSWrapper,
            TanhFunction.apply(
                enc_x, self.coeffs
            )
        )

        return enc_output



from typing import Literal, Union
from numpy.polynomial import Polynomial, Chebyshev
from torchseal.wrapper import CKKSWrapper
from torchseal.function import CrossEntropyLossFunction

import typing
import numpy as np
import torch


class CrossEntropyLoss(torch.nn.Module):
    exp_coeffs: np.ndarray
    inverse_coeffs: np.ndarray
    inverse_iterations: int
    log_coeffs: np.ndarray

    def __init__(
            self,
            exp_start: float,
            exp_stop: float,
            exp_num_of_sample: int,
            exp_degree: int,
            exp_approximation_type: Union[
                Literal["minimax"],
                Literal["least-squares"]
            ],
            inverse_start: float,
            inverse_stop: float,
            inverse_num_of_sample: int,
            inverse_degree: int,
            inverse_approximation_type: Union[
                Literal["minimax"],
                Literal["least-squares"]
            ],
            inverse_iterations: int,
            log_start: float,
            log_stop: float,
            log_num_of_sample: int,
            log_degree: int,
            log_approximation_type: Union[
                Literal["minimax"],
                Literal["least-squares"]
            ]
    ) -> None:
        super(CrossEntropyLoss, self).__init__()

        # Create the polynomial for the exp function
        exp_x = np.linspace(exp_start, exp_stop, exp_num_of_sample)
        exp_y = (lambda x: np.exp(x))(exp_x)

        # Perform the polynomial approximation for the exp function
        self.exp_coeffs = Polynomial.fit(
            exp_x, exp_y, exp_degree
        ).convert(kind=Polynomial).coef if exp_approximation_type == "least-squares" else Chebyshev.fit(exp_x, exp_y, exp_degree).convert(kind=Polynomial).coef

        # Create the polynomial for the inverse function
        inverse_x = np.linspace(
            inverse_start, inverse_stop, inverse_num_of_sample
        )
        inverse_y = (lambda x: 1 / x)(inverse_x)

        # Perform the polynomial approximation for the inverse function
        self.inverse_coeffs = Polynomial.fit(
            inverse_x, inverse_y, inverse_degree
        ).convert(kind=Polynomial).coef if inverse_approximation_type == "least-squares" else Chebyshev.fit(inverse_x, inverse_y, inverse_degree).convert(kind=Polynomial).coef

        # Save the iterations
        self.inverse_iterations = inverse_iterations

        # Create the polynomial for the log function
        log_x = np.linspace(log_start, log_stop, log_num_of_sample)
        log_y = (lambda x: np.log(x))(log_x)

        # Perform the polynomial approximation for the log function
        self.log_coeffs = Polynomial.fit(
            log_x, log_y, log_degree
        ).convert(kind=Polynomial).coef if log_approximation_type == "least-squares" else Chebyshev.fit(log_x, log_y, log_degree).convert(kind=Polynomial).coef

    def forward(self, enc_input: CKKSWrapper, enc_target: CKKSWrapper) -> CKKSWrapper:
        enc_loss = typing.cast(
            CKKSWrapper,
            CrossEntropyLossFunction.apply(
                enc_input, enc_target, self.exp_coeffs, self.inverse_coeffs, self.inverse_iterations, self.log_coeffs
            )
        )

        return enc_loss



from typing import Optional, Tuple
from torchseal.wrapper import CKKSWrapper, CKKSActivationFunctionWithDerivWrapper

import typing
import numpy as np
import torch


class ReLUFunction(torch.autograd.Function):
    @staticmethod
    def forward(ctx: CKKSActivationFunctionWithDerivWrapper, enc_input: CKKSWrapper, coeffs: np.ndarray, deriv_coeffs: np.ndarray) -> CKKSWrapper:
        # Save the ctx for the backward method
        ctx.enc_input = enc_input.clone()
        ctx.deriv_coeffs = deriv_coeffs

        # Apply the sigmoid function to the encrypted input
        enc_output = enc_input.ckks_encrypted_polynomial(coeffs)

        return enc_output

    @staticmethod
    def backward(ctx: CKKSActivationFunctionWithDerivWrapper, enc_grad_output: CKKSWrapper) -> Tuple[Optional[CKKSWrapper], None, None]:
        # Get the saved tensors
        enc_input = ctx.enc_input
        deriv_coeffs = ctx.deriv_coeffs

        # Get the needs_input_grad
        result = typing.cast(Tuple[bool, bool, bool], ctx.needs_input_grad)

        # Initialize the gradients
        enc_grad_input = None

        if result[0]:
            # Do the backward operation
            enc_backward_output = enc_input.ckks_encrypted_polynomial(
                deriv_coeffs
            )

            # Compute the gradients
            enc_grad_input = enc_grad_output.ckks_encrypted_apply_mask(
                enc_backward_output
            )

        return enc_grad_input, None, None



from typing import Optional, Tuple
from torchseal.wrapper import CKKSWrapper, CKKSLinearFunctionWrapper

import typing
import torch


class LinearFunction(torch.autograd.Function):
    @staticmethod
    def forward(ctx: CKKSLinearFunctionWrapper, enc_input: CKKSWrapper, weight: CKKSWrapper, bias: CKKSWrapper, training: bool) -> CKKSWrapper:
        # Save the ctx for the backward method
        ctx.enc_input = enc_input.clone()
        ctx.weight = weight.clone()

        # Apply the linear transformation to the encrypted input
        # If training, apply the linear transformation to the encrypted input using encrypted parameters
        if training:
            enc_output = enc_input.ckks_encrypted_matrix_multiplication(
                weight.ckks_transpose()
            ).ckks_encrypted_addition(bias)

            return enc_output

        # Else, apply the linear transformation to the encrypted input using plaintext parameters
        enc_output = enc_input.ckks_matrix_multiplication(
            weight.plaintext_data.t()
        ).ckks_addition(bias.plaintext_data)

        return enc_output

    @staticmethod
    def backward(ctx: CKKSLinearFunctionWrapper, enc_grad_output: CKKSWrapper) -> Tuple[Optional[CKKSWrapper], Optional[CKKSWrapper], Optional[CKKSWrapper], None]:
        # Get the saved tensors
        enc_weight = ctx.weight
        enc_input = ctx.enc_input

        # Get the needs_input_grad
        result = typing.cast(
            Tuple[bool, bool, bool, bool],
            ctx.needs_input_grad
        )

        # Initialize the gradients
        enc_grad_input = enc_grad_weight = enc_grad_bias = None

        # Compute the gradients
        if result[0]:
            enc_grad_input = enc_grad_output.ckks_encrypted_matrix_multiplication(
                enc_weight
            )
        if result[1]:
            enc_grad_weight = enc_grad_output.ckks_transpose().ckks_encrypted_matrix_multiplication(
                enc_input
            )
        if result[2]:
            enc_grad_bias = enc_grad_output.ckks_sum(axis=0)

        return enc_grad_input, enc_grad_weight, enc_grad_bias, None



from .ckks import CKKSWrapper
from .function import CKKSLinearFunctionWrapper, CKKSPoolingFunctionWrapper, CKKSLossFunctionWrapper, CKKSLossFunctionWithOutputWrapper, CKKSActivationFunctionWrapper, CKKSActivationFunctionWithDerivWrapper



from typing import Optional
from torchseal.wrapper import CKKSWrapper
from torchseal.function import LinearFunction

import typing
import torch
import torchseal


class Linear(torch.nn.Module):
    weight: CKKSWrapper
    bias: CKKSWrapper

    def __init__(self, in_features: int, out_features: int, weight: Optional[CKKSWrapper] = None, bias: Optional[CKKSWrapper] = None):
        super(Linear, self).__init__()

        self.weight = typing.cast(
            CKKSWrapper,
            torch.nn.Parameter(
                torchseal.ckks_wrapper(
                    torch.rand(out_features, in_features),
                    do_encryption=True
                ) if weight is None else weight
            )
        )

        self.bias = typing.cast(
            CKKSWrapper,
            torch.nn.Parameter(
                torchseal.ckks_wrapper(
                    torch.rand(out_features),
                    do_encryption=True
                ) if bias is None else bias
            )
        )

    def forward(self, enc_x: CKKSWrapper) -> CKKSWrapper:
        enc_output = typing.cast(
            CKKSWrapper,
            LinearFunction.apply(
                enc_x, self.weight, self.bias, self.training
            )
        )

        return enc_output

    def train(self, mode=True) -> "Linear":
        if mode:
            # Inplace encrypt the parameters
            self.weight.inplace_encrypt()
            self.bias.inplace_encrypt()
        else:
            # Inplace decrypt the parameters
            self.weight.inplace_decrypt()
            self.bias.inplace_decrypt()

        return super(Linear, self).train(mode)

    def eval(self) -> "Linear":
        return self.train(False)



from typing import Iterator
from torchseal.wrapper import CKKSWrapper

import typing
import torch


class SGD(torch.optim.Optimizer):
    def __init__(self, params: Iterator[torch.nn.Parameter], lr=1e-3) -> None:
        super(SGD, self).__init__(params, defaults={"lr": lr})

    def step(self) -> None:
        for group in self.param_groups:
            for param in group["params"]:
                # If the parameter is not a ckks wrapper or not encrypted, skip it
                if not isinstance(param, CKKSWrapper):
                    continue

                # If the parameter does not require gradients, a ckks wrapper, or not encrypted, skip it
                if param.grad is None or not isinstance(param.grad, CKKSWrapper):
                    continue

                # Both the parameter and its gradient must be encrypted
                assert param.is_encrypted() and param.grad.is_encrypted(
                ), "Only encrypted parameters are allowed."

                # Get the parameter
                lr = typing.cast(float, group["lr"])

                # Delta weight
                delta_weight = param.grad.ckks_apply_scalar(-lr)

                # Update the tensor without creating a new one
                param.ckks_data = param.ckks_data.add(
                    delta_weight.ckks_data
                )



from torchseal.utils import generate_near_zeros

import torch


def create_block_diagonal_matrix(base_matrix: torch.Tensor, num_blocks: int) -> torch.Tensor:
    # Get dimensions of the base matrix
    rows, cols = base_matrix.shape

    # Create a large zero matrix to hold the entire block diagonal matrix
    block_diag_matrix = generate_near_zeros(
        (num_blocks * rows, num_blocks * cols)
    )

    # Fill the block diagonal matrix
    for i in range(num_blocks):
        start_index = i * rows
        block_diag_matrix[
            start_index:start_index + rows, i * cols:i * cols + cols
        ] = base_matrix

    return block_diag_matrix



import torch


def get_sparse_target(target: torch.Tensor, batch_size: int, num_classes: int) -> torch.Tensor:
    # NOTE: This will not to be near_zeros, but zeros because it is applied to plain tensors
    sparse_target = torch.zeros(
        batch_size, num_classes, dtype=torch.long
    ).scatter(1, target.unsqueeze(1), 1)

    return sparse_target



from typing import Sequence, Union
from torchseal.wrapper import CKKSWrapper
from torchseal.state import CKKSState

import torch
import tenseal as ts


def set_context(context: ts.Context) -> None:
    # Get the state of the CKKS
    state = CKKSState()

    # Set the context
    state.context = context


def ckks_wrapper(data: torch.Tensor, do_encryption: bool = False) -> CKKSWrapper:
    # Create the ckks wrapper
    instance = CKKSWrapper(data)

    # If encryption is enabled, encrypt the data
    if do_encryption:
        return instance.encrypt()

    # Else, return the instance without encryption
    return instance


def ckks_zeros(shape: Union[int, Sequence[int]], do_encryption: bool = False) -> CKKSWrapper:
    # Create the ckks wrapper
    instance = CKKSWrapper(torch.zeros(shape))

    # If encryption is enabled, encrypt the data
    if do_encryption:
        return instance.encrypt()

    # Else, return the instance without encryption
    return instance


def ckks_ones(shape: Union[int, Sequence[int]], do_encryption: bool = False) -> CKKSWrapper:
    # Create the ckks wrapper
    instance = CKKSWrapper(torch.ones(shape))

    # If encryption is enabled, encrypt the data
    if do_encryption:
        return instance.encrypt()

    # Else, return the instance without encryption
    return instance



from typing import List

import torch


# NOTE: This is done to handle the case when the batch size is 1 (TenSEAL errors out in this case)
def create_empty_tensors(size: List[int]) -> torch.Tensor:
    if len(size) == 0:
        return torch.zeros(1)

    return torch.zeros(size)



from typing import Sequence, Union

import torch


# This function generate a near-zero tensor with the given shape and scale
# Useful for initializing toeplitz matrices
# NOTE: This is done to solve the possibility of parameter mismatch
def generate_near_zeros(shape: Union[int, Sequence[int]], scale: float = 1e-3) -> torch.Tensor:
    # Randomized with range [0, scale)
    return torch.rand(shape) * scale



from typing import Tuple, Optional
from torchseal.wrapper import CKKSWrapper, CKKSPoolingFunctionWrapper

import typing
import torch


class AvgPool2dFunction(torch.autograd.Function):
    @staticmethod
    def forward(ctx: CKKSPoolingFunctionWrapper, enc_input: CKKSWrapper, weight: torch.Tensor, conv2d_padding_transformation: torch.Tensor, conv2d_inverse_padding_transformation: torch.Tensor) -> CKKSWrapper:
        # Apply the padding transformation to the encrypted input
        # NOTE: This is useless if padding is 0 (we can skip this step if that's the case)
        enc_padding_input = enc_input.ckks_matrix_multiplication(
            conv2d_padding_transformation
        )

        # Save the ctx for the backward method
        ctx.save_for_backward(weight, conv2d_inverse_padding_transformation)

        # Apply the convolution to the encrypted input
        enc_output = enc_padding_input.ckks_matrix_multiplication(weight.t())

        return enc_output

    @staticmethod
    def backward(ctx: CKKSPoolingFunctionWrapper, enc_grad_output: CKKSWrapper) -> Tuple[Optional[CKKSWrapper], None, None, None]:
        # Get the saved tensors
        weight, conv2d_inverse_padding_transformation = typing.cast(
            Tuple[torch.Tensor, torch.Tensor],
            ctx.saved_tensors
        )

        # Get the needs_input_grad
        result = typing.cast(
            Tuple[bool, bool, bool, bool],
            ctx.needs_input_grad
        )

        # Initialize the gradients
        enc_grad_input = None

        # Calculate the gradient for the input
        if result[0]:
            # Calculate the gradients for the input tensor
            enc_padded_grad_input = enc_grad_output.ckks_matrix_multiplication(
                weight
            )

            # Apply the inverse padding transformation to the gradient input
            # NOTE: This is useless if padding is 0 (we can skip this step if that's the case)
            enc_grad_input = enc_padded_grad_input.ckks_matrix_multiplication(
                conv2d_inverse_padding_transformation
            )

        return enc_grad_input, None, None, None



from typing import Literal, Union
from numpy.polynomial import Polynomial, Chebyshev
from torchseal.function import ReLUFunction
from torchseal.wrapper import CKKSWrapper

import typing
import numpy as np
import torch


class ReLU(torch.nn.Module):
    coeffs: np.ndarray
    deriv_coeffs: np.ndarray

    def __init__(
            self,
            start: float,
            stop: float,
            num_of_sample: int,
            degree: int,
            approximation_type: Union[
                Literal["minimax"], Literal["least-squares"]
            ],
            deriv_start: float,
            deriv_stop: float,
            deriv_num_of_sample: int,
            deriv_degree: int,
            deriv_approximation_type: Union[
                Literal["minimax"], Literal["least-squares"]
            ]
    ) -> None:
        super(ReLU, self).__init__()

        # Create the polynomial
        x = np.linspace(start, stop, num_of_sample)
        y = (lambda x: np.maximum(0, x))(x)

        # Perform the polynomial approximation
        self.coeffs = Polynomial.fit(
            x, y, degree
        ).convert(kind=Polynomial).coef if approximation_type == "least-squares" else Chebyshev.fit(x, y, degree).convert(kind=Polynomial).coef

        # Create the polynomial for the derivative
        deriv_x = np.linspace(deriv_start, deriv_stop, deriv_num_of_sample)
        deriv_y = (lambda x: (x > 0) * 1)(deriv_x)

        # Perform the polynomial approximation for the derivative
        self.deriv_coeffs = Polynomial.fit(
            deriv_x, deriv_y, deriv_degree
        ).convert(kind=Polynomial).coef if deriv_approximation_type == "least-squares" else Chebyshev.fit(deriv_x, deriv_y, deriv_degree).convert(kind=Polynomial).coef

    def forward(self, enc_x: CKKSWrapper) -> CKKSWrapper:
        enc_output = typing.cast(
            CKKSWrapper,
            ReLUFunction.apply(
                enc_x, self.coeffs, self.deriv_coeffs
            )
        )

        return enc_output



from typing import Optional, Tuple
from torchseal.wrapper import CKKSWrapper, CKKSActivationFunctionWrapper

import typing
import numpy as np
import torch
import torchseal


class TanhFunction(torch.autograd.Function):
    @staticmethod
    def forward(ctx: CKKSActivationFunctionWrapper, enc_input: CKKSWrapper, coeffs: np.ndarray) -> CKKSWrapper:
        # Apply the sigmoid function to the encrypted input
        enc_output = enc_input.ckks_encrypted_polynomial(coeffs)

        # Save the ctx for the backward method
        ctx.enc_output = enc_output.clone()

        return enc_output

    @staticmethod
    def backward(ctx: CKKSActivationFunctionWrapper, enc_grad_output: CKKSWrapper) -> Tuple[Optional[CKKSWrapper], None]:
        # Get the saved tensors
        enc_output = ctx.enc_output

        # Get the needs_input_grad
        result = typing.cast(Tuple[bool, bool], ctx.needs_input_grad)

        # Initialize the gradients
        enc_grad_input = None

        if result[0]:
            # Compute the gradients (grad_output * (1 - output.pow(2)))
            enc_grad_input = enc_grad_output.ckks_encrypted_apply_mask(
                torchseal.ckks_ones(enc_output.shape, do_encryption=True).ckks_encrypted_addition(
                    enc_output.ckks_encrypted_square().ckks_encrypted_negation()
                )
            )

        return enc_grad_input, None



from torchseal.utils import create_block_diagonal_matrix, generate_near_zeros

import torch


def create_inverse_padding_transformation_matrix(num_channels: int, input_height: int, input_width: int, padding: int) -> torch.Tensor:
    # Add padding to the input size
    padded_input_height = input_height + 2 * padding
    padded_input_width = input_width + 2 * padding

    # Initialize the inverse transformation matrix with zeros
    inverse_transform_matrix = generate_near_zeros(
        (input_height * input_width, padded_input_height * padded_input_width)
    )

    # Fill the inverse transformation matrix to map each padded pixel back to the original image
    for i in range(input_height):
        for j in range(input_width):
            # Original position in the flattened image
            orig_pos = i * input_width + j

            # Corresponding position in the flattened padded image
            padded_pos = (i + padding) * padded_input_width + (j + padding)

            # Update the inverse transformation matrix
            inverse_transform_matrix[orig_pos, padded_pos] = 1

    # Repeat the inverse transformation matrix for each channel
    repeated_inverse_transform_matrix = create_block_diagonal_matrix(
        inverse_transform_matrix, num_channels
    )

    # Transpose the inverse transformation matrix
    transposed_inverse_transform_matrix = repeated_inverse_transform_matrix.t()

    return transposed_inverse_transform_matrix


def create_padding_transformation_matrix(num_channels: int, input_height: int, input_width: int, padding: int) -> torch.Tensor:
    # Add padding to the input size
    padded_input_height = input_height + 2 * padding
    padded_input_width = input_width + 2 * padding

    # Initialize the transformation matrix with zeros
    transform_matrix = generate_near_zeros(
        (padded_input_height * padded_input_width, input_height * input_width)
    )

    # Fill the transformation matrix to map each original pixel to the correct position in the padded image
    for i in range(input_height):
        for j in range(input_width):
            # Original position in the flattened image
            orig_pos = i * input_width + j

            # New position in the flattened padded image
            new_pos = (i + padding) * padded_input_width + (j + padding)

            # Update the transformation matrix
            transform_matrix[new_pos, orig_pos] = 1

    # Repeat the transformation matrix for each channel
    repeated_transform_matrix = create_block_diagonal_matrix(
        transform_matrix, num_channels
    )

    # Transpose the transformation matrix
    transposed_transform_matrix = repeated_transform_matrix.t()

    return transposed_transform_matrix



from typing import Tuple, Optional

from torchseal.wrapper import CKKSWrapper
from torchseal.function import Conv2dFunction
from torchseal.utils import approximate_toeplitz_multiple_channels, create_conv2d_weight_mask, create_conv2d_bias_transformation, create_padding_transformation_matrix, create_inverse_padding_transformation_matrix

import typing
import torch
import torchseal


class Conv2d(torch.nn.Module):
    weight: CKKSWrapper
    bias: CKKSWrapper
    conv2d_padding_transformation: torch.Tensor
    conv2d_inverse_padding_transformation: torch.Tensor
    conv2d_weight_mask: torch.Tensor
    conv2d_bias_transformation: torch.Tensor

    def __init__(self, in_channels: int, out_channels: int, kernel_size: Tuple[int, int], input_size: Tuple[int, int], stride: int = 1, padding: int = 0, weight: Optional[CKKSWrapper] = None, bias: Optional[CKKSWrapper] = None) -> None:
        super(Conv2d, self).__init__()

        # Unpack the kernel size
        kernel_height, kernel_width = kernel_size

        # Unpack the input size
        input_height, input_width = input_size

        # Adjust for padding
        padded_input_height = input_height + 2 * padding
        padded_input_width = input_width + 2 * padding

        # Count the output dimensions
        output_height = (padded_input_height - kernel_height) // stride + 1
        output_width = (padded_input_width - kernel_width) // stride + 1

        # Create the weight and bias
        self.weight = typing.cast(
            CKKSWrapper,
            torch.nn.Parameter(
                torchseal.ckks_wrapper(
                    approximate_toeplitz_multiple_channels(
                        torch.randn(
                            out_channels, in_channels, kernel_height, kernel_width
                        ),
                        (in_channels, input_height, input_width),
                        stride=stride,
                        padding=padding
                    ),
                    do_encryption=True
                ) if weight is None else weight
            )
        )

        self.bias = typing.cast(
            CKKSWrapper,
            torch.nn.Parameter(
                torchseal.ckks_wrapper(
                    torch.repeat_interleave(
                        torch.randn(out_channels),
                        output_height * output_width
                    ),
                    do_encryption=True
                ) if bias is None else bias
            )
        )

        # Create the binary masking for inference
        self.conv2d_padding_transformation = create_padding_transformation_matrix(
            in_channels, input_height, input_width, padding
        )

        self.conv2d_inverse_padding_transformation = create_inverse_padding_transformation_matrix(
            out_channels, input_height, input_width, padding
        )

        # Create the binary masking for training
        self.conv2d_weight_mask = create_conv2d_weight_mask(
            (in_channels, input_height, input_width),
            (out_channels, in_channels, kernel_height, kernel_width),
            stride=stride,
            padding=padding
        )

        self.conv2d_bias_transformation = create_conv2d_bias_transformation(
            repeat=output_height * output_width,
            length=out_channels * output_height * output_width
        )

    def forward(self, enc_x: CKKSWrapper) -> CKKSWrapper:
        enc_output = typing.cast(
            CKKSWrapper,
            Conv2dFunction.apply(
                enc_x, self.weight, self.bias, self.conv2d_padding_transformation, self.conv2d_inverse_padding_transformation, self.conv2d_weight_mask, self.conv2d_bias_transformation, self.training
            )
        )

        return enc_output

    def train(self, mode=True) -> "Conv2d":
        if mode:
            # Inplace encrypt the parameters
            self.weight.inplace_encrypt()
            self.bias.inplace_encrypt()
        else:
            # Inplace decrypt the parameters
            self.weight.inplace_decrypt()
            self.bias.inplace_decrypt()

        return super(Conv2d, self).train(mode)

    def eval(self) -> "Conv2d":
        return self.train(False)



from typing import Optional, Tuple
from torchseal.wrapper import CKKSWrapper, CKKSLossFunctionWithOutputWrapper

import typing
import numpy as np
import torch


class CrossEntropyLossFunction(torch.autograd.Function):
    @staticmethod
    def forward(ctx: CKKSLossFunctionWithOutputWrapper, enc_input: CKKSWrapper, enc_target: CKKSWrapper, exp_coeffs: np.ndarray, inverse_coeffs: np.ndarray, inverse_iterations: int, log_coeffs: np.ndarray) -> CKKSWrapper:
        # Apply softmax to the input to get probabilities
        enc_output = enc_input.ckks_encrypted_softmax(
            exp_coeffs, inverse_coeffs, inverse_iterations
        )

        # Save input and target for backward pass
        ctx.enc_output = enc_output.clone()
        ctx.enc_target = enc_target.clone()

        # Get the batch size
        batch_size, _ = enc_input.shape

        # Calculate the loss
        enc_loss = enc_output.ckks_encrypted_negative_log_likelihood_loss(
            enc_target, log_coeffs, batch_size
        )

        return enc_loss

    @staticmethod
    def backward(ctx: CKKSLossFunctionWithOutputWrapper, enc_grad_output: CKKSWrapper) -> Tuple[Optional[CKKSWrapper], None, None, None, None, None]:
        # Get the saved tensors
        enc_output = ctx.enc_output
        enc_target = ctx.enc_target

        # Unpack the target shape
        batch_size, _ = enc_target.shape

        # Get the needs_input_grad
        result = typing.cast(
            Tuple[
                bool, bool, bool, bool, bool, bool
            ],
            ctx.needs_input_grad
        )

        # Initialize the gradients
        enc_grad_input = None

        if result[0]:
            # Add the subtraction mask to the output (this will be encrypted)
            enc_output_subtracted = enc_output.ckks_encrypted_addition(
                enc_target.ckks_encrypted_negation()
            )

            # Multiply by the grad_output (this will be encrypted)
            enc_grad_input = enc_output_subtracted.ckks_encrypted_apply_mask(
                enc_grad_output
            ).ckks_apply_scalar(1 / batch_size)

        return enc_grad_input, None, None, None, None, None



from torchseal.utils import generate_near_zeros

import torch


def create_average_kernel(n_channels: int, kernel_height: int, kernel_width: int) -> torch.Tensor:
    # Create the initial weight
    initial_weight = generate_near_zeros(
        (n_channels, n_channels, kernel_height, kernel_width)
    )

    # Fill the initial weight with the average pooling kernel
    for i in range(n_channels):
        initial_weight[i, i] = torch.ones(kernel_height, kernel_width).div(
            kernel_height * kernel_width
        )

    return initial_weight



from torchseal.wrapper import CKKSWrapper
from torchseal.function import MSELossFunction

import typing
import torch


class MSELoss(torch.nn.Module):
    def __init__(self) -> None:
        super(MSELoss, self).__init__()

    def forward(self, input: CKKSWrapper, target: CKKSWrapper) -> CKKSWrapper:
        enc_loss = typing.cast(
            CKKSWrapper,
            MSELossFunction.apply(input, target),
        )

        return enc_loss



from .average2d import AvgPool2dFunction
from .conv2d import Conv2dFunction
from .cross_entropy import CrossEntropyLossFunction
from .linear import LinearFunction
from .mse import MSELossFunction
from .relu import ReLUFunction
from .sigmoid import SigmoidFunction
from .square import SquareFunction
from .tanh import TanhFunction



from torchseal.wrapper import CKKSWrapper
from torchseal.function import SquareFunction

import typing
import torch


class Square(torch.nn.Module):
    def __init__(self) -> None:
        super(Square, self).__init__()

    def forward(self, enc_x: CKKSWrapper) -> CKKSWrapper:
        enc_output = typing.cast(
            CKKSWrapper,
            SquareFunction.apply(
                enc_x
            )
        )

        return enc_output



from typing import Optional
from torch.autograd.function import NestedIOFunction
from torchseal.wrapper import CKKSWrapper

import numpy as np


class CKKSLinearFunctionWrapper(NestedIOFunction):
    __enc_input: Optional[CKKSWrapper] = None
    __weight: Optional[CKKSWrapper] = None

    @property
    def enc_input(self) -> CKKSWrapper:
        assert self.__enc_input is not None, "enc_input is not set"

        return self.__enc_input

    @enc_input.setter
    def enc_input(self, enc_input: CKKSWrapper) -> None:
        self.__enc_input = enc_input

    @property
    def weight(self) -> CKKSWrapper:
        assert self.__weight is not None, "weight is not set"

        return self.__weight

    @weight.setter
    def weight(self, weight: CKKSWrapper) -> None:
        self.__weight = weight


class CKKSPoolingFunctionWrapper(NestedIOFunction):
    pass


class CKKSLossFunctionWrapper(NestedIOFunction):
    __enc_input: Optional[CKKSWrapper] = None
    __enc_target: Optional[CKKSWrapper] = None

    @property
    def enc_input(self) -> CKKSWrapper:
        assert self.__enc_input is not None, "enc_input is not set"

        return self.__enc_input

    @enc_input.setter
    def enc_input(self, enc_input: CKKSWrapper) -> None:
        self.__enc_input = enc_input

    @property
    def enc_target(self) -> CKKSWrapper:
        assert self.__enc_target is not None, "enc_target is not set"

        return self.__enc_target

    @enc_target.setter
    def enc_target(self, enc_target: CKKSWrapper) -> None:
        self.__enc_target = enc_target


class CKKSLossFunctionWithOutputWrapper(NestedIOFunction):
    __enc_output: Optional[CKKSWrapper] = None
    __enc_target: Optional[CKKSWrapper] = None

    @property
    def enc_output(self) -> CKKSWrapper:
        assert self.__enc_output is not None, "enc_output is not set"

        return self.__enc_output

    @enc_output.setter
    def enc_output(self, enc_output: CKKSWrapper) -> None:
        self.__enc_output = enc_output

    @property
    def enc_target(self) -> CKKSWrapper:
        assert self.__enc_target is not None, "enc_target is not set"

        return self.__enc_target

    @enc_target.setter
    def enc_target(self, enc_target: CKKSWrapper) -> None:
        self.__enc_target = enc_target


class CKKSActivationFunctionWrapper(NestedIOFunction):
    __enc_output: Optional[CKKSWrapper] = None

    @property
    def enc_output(self) -> CKKSWrapper:
        assert self.__enc_output is not None, "enc_output is not set"

        return self.__enc_output

    @enc_output.setter
    def enc_output(self, enc_output: CKKSWrapper) -> None:
        self.__enc_output = enc_output


class CKKSActivationFunctionWithDerivWrapper(NestedIOFunction):
    __enc_input: Optional[CKKSWrapper] = None
    __deriv_coeffs: Optional[np.ndarray] = None

    @property
    def enc_input(self) -> CKKSWrapper:
        assert self.__enc_input is not None, "enc_input is not set"

        return self.__enc_input

    @enc_input.setter
    def enc_input(self, enc_input: CKKSWrapper) -> None:
        self.__enc_input = enc_input

    @property
    def deriv_coeffs(self) -> np.ndarray:
        assert self.__deriv_coeffs is not None, "deriv_coeffs is not set"

        return self.__deriv_coeffs

    @deriv_coeffs.setter
    def deriv_coeffs(self, deriv_coeffs: np.ndarray) -> None:
        self.__deriv_coeffs = deriv_coeffs



from .near_zeros import generate_near_zeros
from .sparse import get_sparse_target
from .block import create_block_diagonal_matrix
from .toeplitz import precise_toeplitz_multiple_channels, approximate_toeplitz_multiple_channels
from .average2d import create_average_kernel
from .conv2d import create_conv2d_weight_mask, create_conv2d_bias_transformation
from .padding import create_padding_transformation_matrix, create_inverse_padding_transformation_matrix
from .handle_batch_size import create_empty_tensors



from typing import Callable, Dict, Iterable, Optional, Tuple, Type
from torchseal.state import CKKSState
from torchseal.utils import create_empty_tensors


import typing
import numpy as np
import torch
import tenseal as ts
import torchseal


# CKKS Wrapper
# Source: https://colab.research.google.com/drive/1MLeSCMrc6a5Yf_6uAh0bH-IPf5c_BTQf
class CKKSWrapper(torch.Tensor):
    __plaintext_data: torch.Tensor
    __ckks_data: Optional[ts.CKKSTensor] = None

    # Properties
    @property
    def ckks_data(self) -> ts.CKKSTensor:
        assert self.__ckks_data is not None, "CKKS data is not initialized"

        return self.__ckks_data

    # Properties
    @ckks_data.setter
    def ckks_data(self, ckks_data: Optional[ts.CKKSTensor]) -> None:
        self.__ckks_data = ckks_data

    # Properties
    @property
    def plaintext_data(self) -> torch.Tensor:
        return self.__plaintext_data

    # Properties
    @plaintext_data.setter
    def plaintext_data(self, plaintext_data: torch.Tensor) -> None:
        self.__plaintext_data = plaintext_data

    # Predicates
    def is_encrypted(self) -> bool:
        return self.__ckks_data is not None

    # Special methods
    @staticmethod
    def __new__(cls, plaintext_data: torch.Tensor) -> "CKKSWrapper":
        # Create the instance
        instance = torch.Tensor._make_subclass(cls, plaintext_data.to("meta"))

        return instance

    # Special methods
    # NOTE: This will just create unencrypted wrapper
    def __init__(self, plaintext_data: torch.Tensor) -> None:
        # Call the super constructor
        super(CKKSWrapper, self).__init__()

        # Set the plaintext data
        self.plaintext_data = plaintext_data

    # Special methods
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(ENCRYPTED)" if self.is_encrypted() else f"{self.__class__.__name__}(plaintext_data={self.plaintext_data})"

    # Special methods
    # TODO: For improvement, integrate ckks_* function into this method
    @classmethod
    def __torch_dispatch__(cls, func: Callable, _: Iterable[Type], args: Tuple = (), kwargs: Dict = {}) -> "CKKSWrapper":
        # Get the full function name
        full_func_name = f"{func.__module__}.{func.__name__}"

        # Detach function implementation for CKKSWrapper
        if full_func_name == "torch._ops.aten.detach.default" and isinstance(args[0], CKKSWrapper):
            return args[0].detach()

        # Ones like function implementation for CKKSWrapper
        if full_func_name == "torch._ops.aten.ones_like.default" and isinstance(args[0], CKKSWrapper):
            return torchseal.ckks_ones(args[0].shape, do_encryption=args[0].is_encrypted())

        # Else, raise an error
        raise NotImplementedError(
            f"Function {full_func_name} is not implemented for CKKSWrapper"
        )

    # Overridden methods
    # NOTE: Not creating a new memory for the tensor
    def detach(self) -> "CKKSWrapper":
        # Clone the ckks_data
        ckks_data = self.ckks_data

        # Clone the plaintext data
        plaintext_data = self.plaintext_data.detach()

        # Create the new instance
        instance = CKKSWrapper.__new__(CKKSWrapper, plaintext_data)

        # Set the ckks data
        instance.ckks_data = ckks_data

        # Set the plaintext data
        instance.plaintext_data = plaintext_data

        return instance

    # Overridden methods
    # NOTE: Creating a new memory for the tensor
    def clone(self) -> "CKKSWrapper":
        # Clone the ckks_data
        ckks_data = typing.cast(
            ts.CKKSTensor,
            self.ckks_data.copy()
        ) if self.is_encrypted() else None

        # Clone the plaintext data
        plaintext_data = self.plaintext_data.clone()

        # Create the new instance
        instance = CKKSWrapper.__new__(CKKSWrapper, plaintext_data)

        # Set the ckks data
        instance.ckks_data = ckks_data

        # Set the plaintext data
        instance.plaintext_data = plaintext_data

        return instance

    # Overridden methods
    def view_as(self, other: "CKKSWrapper") -> "CKKSWrapper":
        # Resize the ckks_data
        ckks_data = typing.cast(
            ts.CKKSTensor,
            self.ckks_data.reshape(other.ckks_data.shape)
        )

        # Resize the plaintext data
        plaintext_data = self.plaintext_data.view_as(other)

        # Create the new instance
        instance = CKKSWrapper.__new__(CKKSWrapper, plaintext_data)

        # Set the ckks data
        instance.ckks_data = ckks_data

        # Set the plaintext data
        instance.plaintext_data = plaintext_data

        return instance

    # CKKS Operation (with shape change)
    def ckks_matrix_multiplication(self, other: torch.Tensor) -> "CKKSWrapper":
        # Apply the linear transformation to the encrypted input
        ckks_data = self.ckks_data.mm(other.tolist())

        # Create an empty plaintext data tensor
        plaintext_data = create_empty_tensors(ckks_data.shape)

        # Create the new instance
        instance = CKKSWrapper.__new__(CKKSWrapper, plaintext_data)

        # Set the ckks data
        instance.ckks_data = ckks_data

        # Set the plaintext data
        instance.plaintext_data = plaintext_data

        return instance

    # CKKS Operation (with shape change)
    def ckks_encrypted_matrix_multiplication(self, other: "CKKSWrapper") -> "CKKSWrapper":
        # Apply the linear transformation to the encrypted input
        ckks_data = self.ckks_data.mm(other.ckks_data)

        # Create an empty plaintext data tensor
        plaintext_data = create_empty_tensors(ckks_data.shape)

        # Create the new instance
        instance = CKKSWrapper.__new__(CKKSWrapper, plaintext_data)

        # Set the ckks data
        instance.ckks_data = ckks_data

        # Set the plaintext data
        instance.plaintext_data = plaintext_data

        return instance

    # CKKS Operation (with shape change)
    def ckks_apply_transformation(self, transformation: torch.Tensor) -> "CKKSWrapper":
        # Apply the linear transformation to the encrypted input
        ckks_data = self.ckks_data.dot(transformation.tolist())

        # Create an empty plaintext data tensor
        plaintext_data = create_empty_tensors(ckks_data.shape)

        # Create the new instance
        instance = CKKSWrapper.__new__(CKKSWrapper, plaintext_data)

        # Set the ckks data
        instance.ckks_data = ckks_data

        # Set the plaintext data
        instance.plaintext_data = plaintext_data

        return instance

    # CKKS Operation (with shape change)
    def ckks_transpose(self) -> "CKKSWrapper":
        # Apply the transpose function to the encrypted input
        ckks_data = typing.cast(
            ts.CKKSTensor,
            self.ckks_data.transpose()
        )

        # Create an empty plaintext data tensor
        plaintext_data = create_empty_tensors(ckks_data.shape)

        # Create the new instance
        instance = CKKSWrapper.__new__(CKKSWrapper, plaintext_data)

        # Set the ckks data
        instance.ckks_data = ckks_data

        # Set the plaintext data
        instance.plaintext_data = plaintext_data

        return instance

    # CKKS Operation (with shape change)
    def ckks_sum(self, axis: int) -> "CKKSWrapper":
        # Apply the sum function to the encrypted input
        ckks_data = typing.cast(
            ts.CKKSTensor,
            self.ckks_data.sum(axis=axis)
        )

        # Create an empty plaintext data tensor
        plaintext_data = create_empty_tensors(ckks_data.shape)

        # Create the new instance
        instance = CKKSWrapper.__new__(CKKSWrapper, plaintext_data)

        # Set the ckks data
        instance.ckks_data = ckks_data

        # Set the plaintext data
        instance.plaintext_data = plaintext_data

        return instance

    # CKKS Operation
    def ckks_addition(self, other: torch.Tensor) -> "CKKSWrapper":
        # Apply the addition function to the encrypted input
        ckks_data = self.ckks_data.add(other.tolist())

        # Clone the plaintext data
        plaintext_data = self.plaintext_data.clone()

        # Create the new instance
        instance = CKKSWrapper.__new__(CKKSWrapper, plaintext_data)

        # Set the ckks data
        instance.ckks_data = ckks_data

        # Set the plaintext data
        instance.plaintext_data = plaintext_data

        return instance

    # CKKS Operation
    def ckks_encrypted_addition(self, other: "CKKSWrapper") -> "CKKSWrapper":
        # Apply the addition function to the encrypted input
        ckks_data = self.ckks_data.add(other.ckks_data)

        # Clone the plaintext data
        plaintext_data = self.plaintext_data.clone()

        # Create the new instance
        instance = CKKSWrapper.__new__(CKKSWrapper, plaintext_data)

        # Set the ckks data
        instance.ckks_data = ckks_data

        # Set the plaintext data
        instance.plaintext_data = plaintext_data

        return instance

    # CKKS Operation
    def ckks_apply_scalar(self, scalar: float) -> "CKKSWrapper":
        # Apply the scalar function to the encrypted input
        ckks_data = self.ckks_data.mul(scalar)

        # Clone the plaintext data
        plaintext_data = self.plaintext_data.clone()

        # Create the new instance
        instance = CKKSWrapper.__new__(CKKSWrapper, plaintext_data)

        # Set the ckks data
        instance.ckks_data = ckks_data

        # Set the plaintext data
        instance.plaintext_data = plaintext_data

        return instance

    # CKKS Operation
    def ckks_apply_mask(self, mask: torch.Tensor) -> "CKKSWrapper":
        # Apply the mask to the encrypted input
        ckks_data = self.ckks_data.mul(mask.tolist())

        # Clone the plaintext data
        plaintext_data = self.plaintext_data.clone()

        # Create the new instance
        instance = CKKSWrapper.__new__(CKKSWrapper, plaintext_data)

        # Set the ckks data
        instance.ckks_data = ckks_data

        # Set the plaintext data
        instance.plaintext_data = plaintext_data

        return instance

    # CKKS Operation
    def ckks_encrypted_apply_mask(self, mask: "CKKSWrapper") -> "CKKSWrapper":
        # Apply the mask to the encrypted input
        ckks_data = self.ckks_data.mul(mask.ckks_data)

        # Clone the plaintext data
        plaintext_data = self.plaintext_data.clone()

        # Create the new instance
        instance = CKKSWrapper.__new__(CKKSWrapper, plaintext_data)

        # Set the ckks data
        instance.ckks_data = ckks_data

        # Set the plaintext data
        instance.plaintext_data = plaintext_data

        return instance

    # CKKS Operation
    def ckks_encrypted_negation(self) -> "CKKSWrapper":
        # Apply the negation function to the encrypted input
        ckks_data = typing.cast(
            ts.CKKSTensor,
            self.ckks_data.neg()
        )

        # Clone the plaintext data
        plaintext_data = self.plaintext_data.clone()

        # Create the new instance
        instance = CKKSWrapper.__new__(CKKSWrapper, plaintext_data)

        # Set the ckks data
        instance.ckks_data = ckks_data

        # Set the plaintext data
        instance.plaintext_data = plaintext_data

        return instance

    # CKKS Operation
    def ckks_encrypted_square(self) -> "CKKSWrapper":
        # Apply the square function to the encrypted input
        ckks_data = typing.cast(
            ts.CKKSTensor,
            self.ckks_data.square()
        )

        # Clone the plaintext data
        plaintext_data = self.plaintext_data.clone()

        # Create the new instance
        instance = CKKSWrapper.__new__(CKKSWrapper, plaintext_data)

        # Set the ckks data
        instance.ckks_data = ckks_data

        # Set the plaintext data
        instance.plaintext_data = plaintext_data

        return instance

    # CKKS Operation
    def ckks_encrypted_polynomial(self, polyval_coeffs: np.ndarray) -> "CKKSWrapper":
        # Apply the activation function to the encrypted input
        ckks_data = typing.cast(
            ts.CKKSTensor,
            self.ckks_data.polyval(
                polyval_coeffs.tolist()
            )
        )

        # Clone the plaintext data
        plaintext_data = self.plaintext_data.clone()

        # Create the new instance
        instance = CKKSWrapper.__new__(CKKSWrapper, plaintext_data)

        # Set the ckks data
        instance.ckks_data = ckks_data

        # Set the plaintext data
        instance.plaintext_data = plaintext_data

        return instance

    # CKKS Operation
    def ckks_encrypted_softmax(self, exp_coeffs: np.ndarray, inverse_coeffs: np.ndarray, inverse_iterations: int) -> "CKKSWrapper":
        # Apply the exp function to the encrypted input
        act_x = typing.cast(
            ts.CKKSTensor,
            self.ckks_data.polyval(
                exp_coeffs.tolist()
            )
        )

        # Copy the encrypted activation
        act_x_copy = typing.cast(
            ts.CKKSTensor,
            act_x.copy()
        )

        # Apply the sum function to the encrypted input
        sum_x = typing.cast(
            ts.CKKSTensor,
            act_x.sum(axis=1)
        )

        # Apply the multiplicative inverse function to the encrypted input
        inverse_sum_x = typing.cast(
            ts.CKKSTensor,
            sum_x.polyval(
                inverse_coeffs.tolist()
            )
        )

        # Newton-Raphson iteration to refine the inverse
        for _ in range(inverse_iterations):
            prod = sum_x.mul(inverse_sum_x)  # d * x_n

            correction = typing.cast(
                ts.CKKSTensor, prod.neg()
            ).add(
                torch.ones(
                    sum_x.shape
                ).mul(2).tolist()
            )  # 2 - d * x_n

            inverse_sum_x = inverse_sum_x.mul(
                correction
            )  # x_n * (2 - d * x_n)

        # Reshape the inverse sum
        reshaped_inverse_sum_x = typing.cast(
            ts.CKKSTensor,
            inverse_sum_x.reshape(
                [inverse_sum_x.shape[0], 1]
            )
        )

        # Change the inverse sum to a matrix
        binary_matrix = torch.ones(1, act_x_copy.shape[1])
        matrix_inverse_sum = reshaped_inverse_sum_x.mm(binary_matrix.tolist())

        # Apply the division function to the encrypted input
        ckks_data = act_x_copy.mul(matrix_inverse_sum)

        # Clone the plaintext data
        plaintext_data = self.plaintext_data.clone()

        # Create the new instance
        instance = CKKSWrapper.__new__(CKKSWrapper, plaintext_data)

        # Set the ckks data
        instance.ckks_data = ckks_data

        # Set the plaintext data
        instance.plaintext_data = plaintext_data

        return instance

    # CKKS Operation (with shape change)
    def ckks_encrypted_negative_log_likelihood_loss(
        self, enc_target: "CKKSWrapper", log_coeffs: np.ndarray, batch_size: int
    ) -> "CKKSWrapper":
        # Apply log softmax to the output
        enc_log_output = typing.cast(
            ts.CKKSTensor,
            self.ckks_data.polyval(
                log_coeffs.tolist()
            )
        )

        # Calculate the negative log likelihood loss
        enc_log_probs = typing.cast(
            ts.CKKSTensor,
            enc_log_output.mul(enc_target.ckks_data).sum(axis=1)
        )

        # Calculate the loss
        ckks_data = typing.cast(
            ts.CKKSTensor, enc_log_probs.sum(axis=0)
        ).mul(-1 / batch_size)

        # Create an empty plaintext data tensor
        plaintext_data = create_empty_tensors(ckks_data.shape)

        # Create the new instance
        instance = CKKSWrapper.__new__(CKKSWrapper, plaintext_data)

        # Set the ckks data
        instance.ckks_data = ckks_data

        # Set the plaintext data
        instance.plaintext_data = plaintext_data

        return instance

    # Encrypt-Decrypt Operations
    def encrypt(self) -> "CKKSWrapper":
        # Check if the data is already encrypted
        if self.is_encrypted():
            # Copy the encrypted data
            ckks_data = typing.cast(
                ts.CKKSTensor,
                self.ckks_data.copy()
            )

            # Create an empty plaintext data tensor
            plaintext_data = create_empty_tensors(ckks_data.shape)

            # Create the new instance
            instance = CKKSWrapper.__new__(CKKSWrapper, plaintext_data)

            # Set the ckks data
            instance.ckks_data = ckks_data

            # Set the plaintext data
            instance.plaintext_data = plaintext_data

            return instance

        # Get the state of the CKKS
        state = CKKSState()

        # Create the new encrypted tensor
        ckks_data = ts.ckks_tensor(state.context, self.plaintext_data.tolist())

        # Create an empty plaintext data tensor
        plaintext_data = create_empty_tensors(ckks_data.shape)

        # Create the new instance
        instance = CKKSWrapper.__new__(CKKSWrapper, plaintext_data)

        # Set the ckks data
        instance.ckks_data = ckks_data

        # Set the plaintext data
        instance.plaintext_data = plaintext_data

        return instance

    # Encrypt-Decrypt Operations
    def inplace_encrypt(self) -> None:
        # Check if the data is already encrypted
        if self.is_encrypted():
            return

        # Get the state of the CKKS
        state = CKKSState()

        # Create the new encrypted tensor
        self.ckks_data = ts.ckks_tensor(
            state.context, self.plaintext_data.tolist()
        )

        # Create an empty plaintext data tensor
        self.plaintext_data = torch.zeros(self.ckks_data.shape)

    # Encrypt-Decrypt Operations
    def decrypt(self) -> "CKKSWrapper":
        # Check if the data is already decrypted
        if not self.is_encrypted():
            # Set the encrypted data to None
            ckks_data = None

            # Copy the plaintext data
            plaintext_data = self.plaintext_data.clone()

            # Create the new instance
            instance = CKKSWrapper.__new__(CKKSWrapper, plaintext_data)

            # Set the ckks data
            instance.ckks_data = ckks_data

            # Set the plaintext data
            instance.plaintext_data = plaintext_data

            return instance

        # Set the encrypted data to None
        ckks_data = None

        # Decrypt the data
        plaintext_data = torch.tensor(
            self.ckks_data.decrypt().tolist()
        )

        # Create the new instance
        instance = CKKSWrapper.__new__(CKKSWrapper, plaintext_data)

        # Set the ckks data
        instance.ckks_data = ckks_data

        # Set the plaintext data
        instance.plaintext_data = plaintext_data

        return instance

    # Encrypt-Decrypt Operations
    def inplace_decrypt(self) -> None:
        # Check if the data is already decrypted
        if not self.is_encrypted():
            return

        # Decrypt the data
        self.plaintext_data = torch.tensor(
            self.ckks_data.decrypt().tolist()
        )

        # Set the encrypted data to None
        self.ckks_data = None



from .average2d import AvgPool2d
from .conv2d import Conv2d
from .cross_entropy import CrossEntropyLoss
from .linear import Linear
from .mse import MSELoss
from .relu import ReLU
from .sigmoid import Sigmoid
from .square import Square
from .tanh import Tanh



from typing import Optional, Tuple
from torchseal.wrapper import CKKSWrapper, CKKSLossFunctionWrapper

import typing
import torch


class MSELossFunction(torch.autograd.Function):
    @staticmethod
    def forward(ctx: CKKSLossFunctionWrapper, enc_input: CKKSWrapper, enc_target: CKKSWrapper) -> CKKSWrapper:
        # Save input and target for backward pass
        ctx.enc_input = enc_input.clone()
        ctx.enc_target = enc_target.clone()

        # Calculate the difference between the input and the target
        enc_diff = enc_input.ckks_encrypted_addition(
            enc_target.ckks_encrypted_negation()
        )

        # Calculate the square of the difference
        enc_diff_squared = enc_diff.ckks_encrypted_square()

        # Sum the squared differences
        enc_diff_squared_sum = enc_diff_squared.ckks_sum(
            axis=1
        ).ckks_sum(
            axis=0
        )

        # Get the number of elements in the input
        batch_size, input_dim = enc_input.shape
        numel = batch_size * input_dim

        # Calculate the mean of the loss
        enc_loss = enc_diff_squared_sum.ckks_apply_scalar(1 / numel)

        return enc_loss

    @staticmethod
    def backward(ctx: CKKSLossFunctionWrapper, enc_grad_output: CKKSWrapper) -> Tuple[Optional[CKKSWrapper], None]:
        # Get the saved tensors
        enc_input = ctx.enc_input
        enc_target = ctx.enc_target

        # Unpack the input shape
        batch_size, input_dim = enc_input.shape
        numel = batch_size * input_dim

        # Get the needs_input_grad
        result = typing.cast(
            Tuple[
                bool, bool
            ],
            ctx.needs_input_grad
        )

        # Initialize the gradients
        enc_grad_input = None

        if result[0]:
            # Calculate the subtraction mask
            enc_input_subtracted = enc_input.ckks_encrypted_addition(
                enc_target.ckks_encrypted_negation()
            ).ckks_apply_scalar(2 / numel)

            # Multiply by the grad_output
            enc_grad_input = enc_input_subtracted.ckks_encrypted_apply_mask(
                enc_grad_output
            )

        return enc_grad_input, None



from typing import Optional, Tuple
from torchseal.wrapper import CKKSWrapper, CKKSLinearFunctionWrapper

import typing
import torch


class SquareFunction(torch.autograd.Function):
    @staticmethod
    def forward(ctx: CKKSLinearFunctionWrapper, enc_input: CKKSWrapper) -> CKKSWrapper:
        # Save the ctx for the backward method
        ctx.enc_input = enc_input.clone()

        # Apply square function to the encrypted input (x ** 2)
        enc_output = enc_input.ckks_encrypted_square()

        return enc_output

    @staticmethod
    def backward(ctx: CKKSLinearFunctionWrapper, enc_grad_output: CKKSWrapper) -> Tuple[Optional[CKKSWrapper]]:
        # Get the saved tensors
        enc_input = ctx.enc_input

        # Get the needs_input_grad
        result = typing.cast(Tuple[bool], ctx.needs_input_grad)

        # Initialize the gradients
        enc_grad_input = None

        if result[0]:
            # Do the backward operation
            enc_backward_output = enc_input.ckks_apply_scalar(2)

            # Compute the gradients
            enc_grad_input = enc_grad_output.ckks_encrypted_apply_mask(
                enc_backward_output
            )

        return enc_grad_input,



from typing import Tuple
from torchseal.utils import generate_near_zeros

import torch


# Source: https://stackoverflow.com/questions/68896578/pytorchs-torch-as-strided-with-negative-strides-for-making-a-toeplitz-matrix/68899386#68899386
def precise_toeplitz(c: torch.Tensor, r: torch.Tensor) -> torch.Tensor:
    vals = torch.cat((r, c[1:].flip(0)))
    shape = len(c), len(r)
    i, j = torch.ones(*shape).nonzero().T

    return vals[j - i].view(*shape)


# Source: https://stackoverflow.com/questions/56702873/is-there-an-function-in-pytorch-for-converting-convolutions-to-fully-connected-n
def precise_toeplitz_one_channel(kernel: torch.Tensor, input_size: Tuple[int, int], stride: int = 1, padding: int = 0) -> torch.Tensor:
    kernel_height, kernel_width = kernel.shape
    input_height, input_width = input_size

    # Adjust the input dimensions based on padding
    padded_input_height = input_height + 2 * padding
    padded_input_width = input_width + 2 * padding

    # Calculate the output dimensions considering stride and padding
    output_height = ((padded_input_height - kernel_height) // stride) + 1

    # Stack and reshape the matrices to form the final weight matrix
    toeplitz_matrices = torch.stack([
        # Construct 1D convolution toeplitz matrices for each row of the kernel considering stride
        precise_toeplitz(
            torch.cat(
                [
                    kernel[r, 0:1],
                    torch.zeros(
                        padded_input_height - kernel_height
                    )
                ]
            ),
            torch.cat(
                [kernel[r, :], torch.zeros(
                    padded_input_width - kernel_width)]
            )
        )[::stride, :][:output_height]
        for r in range(kernel_height)
    ])

    # Calculate the number of blocks and their sizes for constructing the final matrix
    num_blocks_height = output_height
    block_height, block_width = toeplitz_matrices[0].shape

    # Initialize the final weight matrix with zeros
    weight_matrix = torch.zeros(
        (num_blocks_height * block_height, padded_input_width * padded_input_height)
    )

    # Fill in the blocks for the final weight matrix
    for i in range(kernel_height):
        for j in range(output_height):
            start_row = j * block_height
            end_row = start_row + block_height
            start_col = (i + j * stride) * padded_input_width
            end_col = start_col + block_width
            weight_matrix[
                start_row:end_row, start_col:end_col
            ] = toeplitz_matrices[i]

    return weight_matrix


# Source: https://stackoverflow.com/questions/56702873/is-there-an-function-in-pytorch-for-converting-convolutions-to-fully-connected-n
def precise_toeplitz_multiple_channels(kernel: torch.Tensor, input_size_with_channel: Tuple[int, int, int], stride: int = 1, padding: int = 0) -> torch.Tensor:
    # Get the shapes
    kernel_out_channel, _, kernel_height, kernel_width = kernel.shape
    input_in_channel, input_height, input_width = input_size_with_channel

    # Adjust for padding
    padded_input_height = input_height + 2 * padding
    padded_input_width = input_width + 2 * padding

    # Calculate the output size (with padding and stride)
    output_height = (padded_input_height - kernel_height) // stride + 1
    output_width = (padded_input_width - kernel_width) // stride + 1

    # Initialize the output tensor
    weight_convolutions = torch.zeros(
        (
            kernel_out_channel,
            output_height * output_width,
            input_in_channel,
            padded_input_height * padded_input_width
        )
    )

    # Fill the output tensor
    for i, kernel_output in enumerate(kernel):
        for j, kernel_input in enumerate(kernel_output):
            weight_convolutions[i, :, j, :] = precise_toeplitz_one_channel(
                kernel_input, input_size_with_channel[1:], padding=padding, stride=stride
            )

    # Reshape the output tensor
    weight_convolutions = weight_convolutions.view(
        kernel_out_channel * output_height * output_width,
        input_in_channel * padded_input_height * padded_input_width
    )

    return weight_convolutions


# Source: https://stackoverflow.com/questions/68896578/pytorchs-torch-as-strided-with-negative-strides-for-making-a-toeplitz-matrix/68899386#68899386
def approximate_toeplitz(c: torch.Tensor, r: torch.Tensor) -> torch.Tensor:
    vals = torch.cat((r, c[1:].flip(0)))
    shape = len(c), len(r)
    i, j = torch.ones(*shape).nonzero().T

    return vals[j - i].view(*shape)


# Source: https://stackoverflow.com/questions/56702873/is-there-an-function-in-pytorch-for-converting-convolutions-to-fully-connected-n
def approximate_toeplitz_one_channel(kernel: torch.Tensor, input_size: Tuple[int, int], stride: int = 1, padding: int = 0) -> torch.Tensor:
    kernel_height, kernel_width = kernel.shape
    input_height, input_width = input_size

    # Adjust the input dimensions based on padding
    padded_input_height = input_height + 2 * padding
    padded_input_width = input_width + 2 * padding

    # Calculate the output dimensions considering stride and padding
    output_height = ((padded_input_height - kernel_height) // stride) + 1

    # Stack and reshape the matrices to form the final weight matrix
    toeplitz_matrices = torch.stack([
        # Construct 1D convolution toeplitz matrices for each row of the kernel considering stride
        approximate_toeplitz(
            torch.cat(
                [
                    kernel[r, 0:1],
                    generate_near_zeros(
                        padded_input_height - kernel_height
                    )
                ]
            ),
            torch.cat(
                [kernel[r, :], generate_near_zeros(
                    padded_input_width - kernel_width)]
            )
        )[::stride, :][:output_height]
        for r in range(kernel_height)
    ])

    # Calculate the number of blocks and their sizes for constructing the final matrix
    num_blocks_height = output_height
    block_height, block_width = toeplitz_matrices[0].shape

    # Initialize the final weight matrix with zeros
    weight_matrix = generate_near_zeros(
        (num_blocks_height * block_height, padded_input_width * padded_input_height)
    )

    # Fill in the blocks for the final weight matrix
    for i in range(kernel_height):
        for j in range(output_height):
            start_row = j * block_height
            end_row = start_row + block_height
            start_col = (i + j * stride) * padded_input_width
            end_col = start_col + block_width
            weight_matrix[
                start_row:end_row, start_col:end_col
            ] = toeplitz_matrices[i]

    return weight_matrix


# Source: https://stackoverflow.com/questions/56702873/is-there-an-function-in-pytorch-for-converting-convolutions-to-fully-connected-n
def approximate_toeplitz_multiple_channels(kernel: torch.Tensor, input_size_with_channel: Tuple[int, int, int], stride: int = 1, padding: int = 0) -> torch.Tensor:
    # Get the shapes
    kernel_out_channel, _, kernel_height, kernel_width = kernel.shape
    input_in_channel, input_height, input_width = input_size_with_channel

    # Adjust for padding
    padded_input_height = input_height + 2 * padding
    padded_input_width = input_width + 2 * padding

    # Calculate the output size (with padding and stride)
    output_height = (padded_input_height - kernel_height) // stride + 1
    output_width = (padded_input_width - kernel_width) // stride + 1

    # Initialize the output tensor
    weight_convolutions = generate_near_zeros(
        (
            kernel_out_channel,
            output_height * output_width,
            input_in_channel,
            padded_input_height * padded_input_width
        )
    )

    # Fill the output tensor
    for i, kernel_output in enumerate(kernel):
        for j, kernel_input in enumerate(kernel_output):
            weight_convolutions[i, :, j, :] = approximate_toeplitz_one_channel(
                kernel_input, input_size_with_channel[1:], padding=padding, stride=stride
            )

    # Reshape the output tensor
    weight_convolutions = weight_convolutions.view(
        kernel_out_channel * output_height * output_width,
        input_in_channel * padded_input_height * padded_input_width
    )

    return weight_convolutions
